#include <jabberd.h>
#define IRC_T_VERSION "0.9"

#define IDX_IRC
#ifdef IDX_IRC
#define IDX_IRC_PASSWD "irc2int4idx"
#endif

typedef struct irctrans_data_st
{
    instance i;
    mtq q;
    xht sessions;
    xdbcache xc;
    xmlnode browse;
    xmlnode cfg;
    char *default_server;
    int default_port;
    time_t start;
    int rand_room; /* for random name generation of channels */
    char *note_leave, *note_join, *note_rename, *note_topic; /* notices */
} _irctrans_data, *irctrans_data;

extern irctrans_data ircdata;

/* represent an user in a channel */
/* all the users in a channel are "stored" in such a structure */
typedef struct irc_user_st {
    pool p;
    char *user;           /* nickname */
    jid id;               /* cf nick_id */
    int ops;              /* IRC ops rights - unused */
    int voice;            /* IRC voice rights - unused */
    void *data;           /* to store additional data, usually a channel struct */
} _user, *user;

typedef enum { chan_JOINING, chan_ACTIVE, chan_INACTIVE } chan_state;
typedef struct chan_struct {
    pool p;
    chan_state state;
    char *name;           /* channel name */
    char *topic;          /* channel topic */
    char *nick_id;        /* user id in channel, eg #jabber@irc.myserver.com/key */
    char *id;             /* channel id, eg #jabber@irc.myserver.com */
    xht users;            /* users table */
    void *data;           /* to store additional data, usually a session struct */
} _chans, *chans;

/* whether we are entering an existing room or creating a new one */
typedef enum { sess_ENTER, sess_CREATE } sess_state;

/* a session is associated to an user logged in an IRC server */
typedef struct irc_sessions_struct
{
    pool p,pbuffer;
    sess_state state;     /* hmmm, put it in chan_struct ? */
    mio m;                /* socket to IRC server */
    char *id;             /* user_full_jid/irc_server */
    jid j;                /* user remote jid */
    int port;             /* unused ... but it should be used ! */
    char *nick;           /* user nickname on this server */
    char *server;         /* IRC server user is logged into */
    char *buffer;         /* to store received IRC commands */
    xht channels;         /* user channels */
    void *aux1;           
} _irc_session, *irc_session;

void irc_parser(void *arg); /* main jpacket parser */

char *irc_get_server(pool p, char *input); 
char *irc_get_channel(pool p, char *input); 
char *irc_get_port(pool p, char *input);

void irc_parse_phase_zero(mio m, int flag, void *arg, char *buffer, int bufsz);
void irc_parse_phase_one(mio m, int flag, void *arg, char *buffer, int bufsz);
void irc_check_iq_get(mio m, int flag, void *arg, char *buffer, int bufsz);
void irc_check_browse_get(mio m, int flag, void *arg, char *buffer, int bufsz);

void irc_strip_mirc_codes(char *input);
unsigned char *UTF8_to_str(pool p,unsigned char *in);
char *str_to_UTF8(pool p,unsigned char *in);
char *make_lowercase(char *input);

