/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>

/* global variable - contains the global configuration of the transport */
irctrans_data ircdata;

result irc_phandler_receive(instance id, dpacket dp, void *arg)
{
    jpacket jp;

    log_debug(ZONE, "[irc_phandler_receive] got %s", xmlnode2str(dp->x));

    /* strip the route off */
    if(dp->type == p_ROUTE)
    {
        jutil_error(dp->x, TERROR_NOTALLOWED); /* we do not accept routes */
        deliver(dpacket_new(dp->x), NULL);
        return r_DONE;
    }

    jp = jpacket_new(dp->x);

    if(jp == NULL || jp->type == JPACKET_UNKNOWN)
    {
        xmlnode_free(jp->x);
        return r_DONE;
    }

    log_debug(ZONE, "[irc_phandler_receive] handing off to mtq");
    mtq_send(ircdata->q, jp->p, irc_parser, (void*)jp);
    return r_DONE;
}

void _irc_trans_stop(xht sessions, const char *key, void *val, void *arg)
{
    xmlnode m = xmlnode_new_tag("message");
    irc_session s = (irc_session)val;
    irctrans_data ircdata = (irctrans_data)arg;

    xmlnode_put_attrib(m, "type", "error");
    xmlnode_put_attrib(m, "from", jid_full(s->j));
    xmlnode_put_attrib(m, "to", ircdata->i->id);

    jutil_error(m, TERROR_EXTERNAL);
    xmlnode_insert_cdata(xmlnode_insert_tag(m, "body"), "IRC Transport is shutting down", -1);

    deliver(dpacket_new(m), NULL);

    mio_close(s->m);
}

/* used to exit the server and close down all session gracefully */
void irc_trans_stop(void *arg)
{
    xhash_walk(ircdata->sessions, _irc_trans_stop, (void*)ircdata);
    xhash_free(ircdata->sessions);
}

int handle_config(xmlnode conf)
{
    xmlnode cur;
    ircdata->default_port = 0;

    for(cur = xmlnode_get_firstchild(conf); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if (xmlnode_get_type(cur) != NTYPE_TAG)
            continue;

        if (j_strcmp(xmlnode_get_name(cur), "default") == 0)
        {
            ircdata->default_server = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "default/server"));
            ircdata->default_port   = j_atoi(xmlnode_get_tag_data(conf, "default/port"),6667);
	    log_debug(ZONE,"[handle_config] got defaut server/port : %s/%d", ircdata->default_server, ircdata->default_port);
        }

	if (j_strcmp(xmlnode_get_name(cur), "notice") == 0) {
	    ircdata->note_join = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "notice/join"));
	    ircdata->note_leave = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "notice/leave"));
	    ircdata->note_rename = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "notice/rename"));
	    ircdata->note_topic = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "notice/topic"));
	}
    }

    if(ircdata->default_port == 0)
      ircdata->default_port = 6667;
    if(ircdata->default_server == NULL)
      ircdata->default_server = pstrdup(ircdata->i->p, "irc.linux.com");
    return 0;
}


void __irc_debug_print_users(xht users, const char *key, void *val, void *arg)
{
    user u = (user)val;
    log_debug("irc_transport", "----user[%X] name: %s id; %s ops/void: %d/%d", val, u->user, jid_full(u->id), u->ops, u->voice);
}

void __irc_debug_print_channels(xht channels, const char *key, void *val, void *arg)
{
    chans c = (chans)val;
    log_debug("irc_transport", "--Channel[%X] name: %s nick_id: %s id: %s state: %d", val, c->name, c->nick_id, c->id, c->state);
    xhash_walk(c->users, __irc_debug_print_users, arg);
}

void __irc_debug_print_sessions(xht sessions, const char *key, void *val, void *arg)
{
    irc_session s = (irc_session)val;
    log_debug("irc_transport", "Session[%X] id: %s j: %s nick: %s server: %s", val, s->id, jid_full(s->j), s->nick, s->server);
    xhash_walk(s->channels, __irc_debug_print_channels, arg);
}

result _irc_debug_loop(void *arg)
{
    log_debug("irc_transport", "Printint Sessions:");    
    xhash_walk(ircdata->sessions, __irc_debug_print_sessions, arg);
    return r_DONE;
}

/* load up the irc transport from jabberd */
void irc_transport(instance i, xmlnode x)
{
    xmlnode config;

    log_notice(i->id, "IRC Transport starting up...");
    log_debug("irc_transport", "[%s] starting up", ZONE);

    ircdata            = pmalloco(i->p, sizeof(_irctrans_data));
    ircdata->i         = i;
    ircdata->xc        = xdb_cache(i);
    ircdata->q         = mtq_new(i->p);
    ircdata->cfg       = xdb_get(ircdata->xc, jid_new(xmlnode_pool(x), "config@-internal"), "jabber:config:irc");
    ircdata->sessions  = xhash_new(101);
    ircdata->browse    = xmlnode_new_tag("query");
    ircdata->start     = time(NULL);
    ircdata->rand_room = 0;

    handle_config(ircdata->cfg);

    xmlnode_put_attrib(ircdata->browse, "xmlns", "jabber:iq:browse");
    xmlnode_put_attrib(xmlnode_insert_tag(ircdata->browse, "service"), "type", "irc");
    xmlnode_put_attrib(xmlnode_get_tag(ircdata->browse, "service"), "jid", i->id);
    xmlnode_put_attrib(xmlnode_get_tag(ircdata->browse, "service"), "name", xmlnode_get_tag_data(ircdata->cfg, "vCard/FN"));
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_VERSION, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_BROWSE, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_TIME, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_LAST, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_VCARD, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_CONFERENCE, -1);

    log_debug("irc_transport", "[%s] created browse packet: %s", ZONE, xmlnode2str(ircdata->browse));
    log_debug("irc_transport", "[%s] IRC Transport Loading Configuration: %s", ZONE, xmlnode2str(config));

    register_phandler(i, o_DELIVER, irc_phandler_receive, (void*) NULL);

    if(xmlnode_get_tag(ircdata->cfg, "update") != NULL)
    {
        xmlnode x;
        spool sp;
        log_debug("irc_transport","[%s] sending out update query", ZONE);
        x = xmlnode_new_tag("presence");
        sp = spool_new(xmlnode_pool(x));
        xmlnode_put_attrib(x, "to", spools(xmlnode_pool(x),"964735456@", xmlnode_get_tag_data(ircdata->cfg, "update/host"), "/", IRC_T_VERSION, xmlnode_pool(x)));
        xmlnode_put_attrib(x, "from", ircdata->i->id);
        deliver(dpacket_new(x), i);
    }

    pool_cleanup(i->p, irc_trans_stop, (void*) NULL);

    /* Uncomment to print out hash data every 5 seconds */
    /* register_beat(5, _irc_debug_loop, NULL); */
}
