/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>
#include <sys/utsname.h>


irc_session irc_new_session(jpacket jp)
{
    irc_session new;
    pool        p;

    log_debug(ZONE, "[irc_new_session] creating a new session for %s", jid_full(jp->from));

    p             = pool_new();
    new           = pmalloco(p, sizeof(_irc_session));
    new->j        = jid_new(p, jid_full(jp->from));
    jp->from->next = (jid)ircdata;
    new->server   = irc_get_server(p, jid_full(jp->to));
    new->id       = spools(p, jid_full(jp->from), "/", new->server, p);
    new->p        = p;
    new->channels = xhash_new(101);
    new->nick     = pstrdup(p, jp->from->user);

    return new;
}

/* parse requests sent directly to the IRC transport */
void irc_parse_server(jpacket jp)
{
    if ((jp->type != JPACKET_IQ) || (jpacket_subtype(jp) != JPACKET__GET))
    {
	jutil_error(jp->x, TERROR_BAD);
	deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_TIME))
    {
        time_t t;
        char   *str;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_TIME);
        jpacket_reset(jp);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"utc"),jutil_timestamp(),-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"tz"),tzname[0],-1);
        
        /* create nice display time */
        t = time(NULL);
        str = ctime(&t);
        str[strlen(str) - 1] = '\0'; /* cut off newline */
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"display"),str,-1);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_VERSION))
    {
        struct utsname un;
        xmlnode x;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_VERSION);
        jpacket_reset(jp);
 
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"name"),"IRC Transport",-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"version"),IRC_T_VERSION,-1);
 
        uname(&un);
        x = xmlnode_insert_tag(jp->iq,"os");
        xmlnode_insert_cdata(x,un.sysname,-1);
        xmlnode_insert_cdata(x," ",1);
        xmlnode_insert_cdata(x,un.release,-1);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_BROWSE))
    {
        irc_session s;
	xmlnode iq;

	/* browse request to a server or a channel */
        if(jp->to->resource == NULL)
        {
	  /* create a temp session to get info */
	  s = irc_new_session(jp);
	}
	/* browse request to a specific user */
        else
	{
	    /* use user session */
	    /* XXX : send the WHOIS request here and do not create a new conn */
	    /*       and handles the RPL_WHOREPLY in parse_phase_two ... */
	    /* by BO, 2001/10/10 */
	    s = (irc_session) xhash_get(ircdata->sessions, spools(jp->p, jid_full(jp->from), "/", irc_get_server(jp->p, jid_full(jp->to)), jp->p));
	    if (s == NULL) {
		log_error(ZONE,"[irc_parse_server] User has no session !");
		jutil_iqresult(jp->x);
		jutil_error(jp->x,TERROR_REGISTER);
		deliver(dpacket_new(jp->x),NULL);
		return;
	    }
	}

	jutil_iqresult(jp->x);
	iq = xmlnode_insert_tag(jp->x,"query");
	xmlnode_put_attrib(iq,"xmlns",NS_BROWSE);
	jpacket_reset(jp);
    
	jp->aux1 = (void*) s;
	log_debug(ZONE, "[irc_parse_server] calling mio_connect(%s, %d)", s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), ircdata->default_port));
	mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), ircdata->default_port), irc_check_browse_get, (void*)jp, 15, MIO_CONNECT_RAW);
	return;
    }
    else if(NSCHECK(jp->iq,NS_LAST))
    {
        time_t start;
        char nstr[20];
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_LAST);
        jpacket_reset(jp);
       
        start = time(NULL) - ircdata->start;
        sprintf(nstr,"%d",(int) start);
        xmlnode_put_attrib(jp->iq,"seconds",nstr);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_VCARD))
    {
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"vCard"),"xmlns",NS_VCARD);
        jpacket_reset(jp);

        xmlnode_insert_node(jp->iq,xmlnode_get_firstchild(xmlnode_get_tag(ircdata->cfg,"vCard")));
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x),NULL);
}

void irc_parser(void *arg)
{
    /* incoming packets from the jabber user start here */
    jpacket jp = (jpacket)arg;
    irc_session s;

    if(jpacket_subtype(jp) == JPACKET__ERROR)
    {
        /* we don't care about returned errors yet.. */
        /* XXX depending on the error.. should this drop their sessions? */
        xmlnode_free(jp->x);
        return;
    }

    /* hide the ircdata in the jpacket XXX why not use aux1? */
    jp->from->next = (jid)ircdata;

    /* branch off server packets */
    if (((jp->to->user == NULL) && (!NSCHECK(jp->iq, NS_CONFERENCE)))
	|| (NSCHECK(jp->iq,NS_BROWSE)))
    {
        irc_parse_server(jp);
        return;
    }
    
    /* see if this user has an existing session */
    /* format is: tsbandit@jabber.org/resource/irc.linux.com */
    s = (irc_session) xhash_get(ircdata->sessions, spools(jp->p, jid_full(jp->from), "/", irc_get_server(jp->p, jid_full(jp->to)), jp->p));
    if (s != NULL)
	log_debug(ZONE,"[irc_parser] found a session for user %s", jid_full(jp->from));

    /* handles request on jabber:iq:conference namespace */
    if (NSCHECK(jp->iq, NS_CONFERENCE))
    {
	if (jp->type != JPACKET_IQ)
	{
	    log_warn(ZONE,"[irc_parser] conference ns request badly formed");
	    jutil_error(jp->x, TERROR_BAD);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	/* enter/create get requests */
	if (((strncmp(xmlnode_get_name(jp->iq),"enter",5) == 0) || (strncmp(xmlnode_get_name(jp->iq),"create",6) == 0))
	    && (jpacket_subtype(jp) == JPACKET__GET))
	{
	    /* create a session, to handle the irc server lookup */
	    /* except if user already a session on this server */
	    if (s == NULL) {
		s = irc_new_session(jp);
		/* insert this session into the xhash */
		xhash_put(ircdata->sessions, s->id, (void*)s);

		log_debug(ZONE, "[irc_parser] Initial session started for user %s for channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);
	    } else {
		log_debug(ZONE, "[irc_parser] Using previously created session for user %s for channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);
	    }

	    jutil_iqresult(jp->x);
	    /* store phase we're in */
	    if (strncmp(xmlnode_get_name(jp->iq),"enter",5) == 0) {
	      xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"enter"),"xmlns",NS_CONFERENCE);
	      s->state = sess_ENTER;

	      jpacket_reset(jp);
	      xmlnode_insert_tag(jp->iq, "nickname");
	      jp->aux1 = (void*) s;
	      log_debug(ZONE, "[irc_parser] calling mio_connect(%s, %d)", s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), ircdata->default_port));
	      /* log into the irc server, to check if we need a password 
	       * for this channel 
	       * this function will deliver the jpacket response as well.. 
	       * we don't handle this request from here any longer */
	      mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), ircdata->default_port), irc_check_iq_get, (void*)jp, 15, MIO_CONNECT_RAW);
	      return;
	    } else {
		xmlnode temp,create;

		create = xmlnode_insert_tag(jp->x,"create");
		xmlnode_put_attrib(create,"xmlns",NS_CONFERENCE);
		s->state = sess_CREATE;

		/* ask for info to create a room */
		xmlnode_insert_tag(create,"address");
		xmlnode_insert_tag(create,"nickname");
		xmlnode_insert_tag(create,"password");
		temp = xmlnode_new_tag("metainfo");
		xmlnode_insert_tag(temp,"name");
		xmlnode_insert_tag(temp,"desc");
		xmlnode_insert_node(create,temp);
		deliver(dpacket_new(jp->x), NULL);
		return;
	    }
	}

	/* enter/create set requests */
	if (((strncmp(xmlnode_get_name(jp->iq),"enter",5) == 0) || (strncmp(xmlnode_get_name(jp->iq),"create",6) == 0))
	    && (jpacket_subtype(jp) == JPACKET__SET))
	{
	    char *chan = make_lowercase(irc_get_channel(jp->p, jid_full(jp->to)));
	    chans channel = NULL;

	    /* user did not make an get request before */
	    if (s == NULL) {
		log_debug(ZONE,"[irc_parser] no session for %s, creating a new one", jid_full(jp->from));
		/* start a new irc session */
		s = irc_new_session(jp);
		/* insert this session into the xhash */
		xhash_put(ircdata->sessions, s->id, (void*)s);
	    }

	    /* s->m isn't set until the session arrives in phase two */
	    if(s->m == NULL) { 
		/* we don't have a connection to the server yet */
		jp->aux1 = (void*)s;
		log_debug(ZONE, "[irc_parser] Bootstrapping Connection for user %s: channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);
		/* this will connect, and add the channel to the xhash */
		mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->to)), ircdata->default_port), irc_parse_phase_zero, (void*)jp, 10, MIO_CONNECT_RAW);
		return;
	    }

	    channel = xhash_get(s->channels, chan);
	    if (channel == NULL) { 
		/* startup a new channel for this session (which is in phase two) */
		log_debug(ZONE,"[irc_parser] starting a new channel for user %s", jid_full(jp->from));
                jp->aux1 = (void*)s;
                s->aux1 = (void*)jp;
		if (s->state == sess_ENTER) {
		    mio_write(s->m, NULL, spools(jp->p, "JOIN ", irc_get_channel(jp->p, jid_full(jp->to)), " ", xmlnode_get_tag_data(jp->iq, "password"), "\n", jp->p), -1);
		} else if (s->state == sess_CREATE) {
		    /* retrieve or create room name */
		    /* send the MODE cmd to be sure channel does not exist 
		     * we will get the results in irc_parse_phase_two */
		    /* XXX : this code is exactly the same as in
		     *       irc_parse_phase_two
		     *       try to do it better */
		    char *address, *chan_name;

		    /* did user specify a room name ? */
		    if ((address = xmlnode_get_tag_data(jp->iq, "address")) == NULL) {
			/* randomly set a channel name */
			ircdata->rand_room++;
			chan_name = pmalloc(s->p,strlen("#parisestmagique")+5);
			sprintf(chan_name,"%s%d","#parisestmagique",ircdata->rand_room);
			log_debug(ZONE,"[irc_parser] No room name, randomly setting : %s", chan_name);
		    } else {
			char *ptr;
			/* remove the @ if it was specified */
			ptr = strchr(address,'@');
			if (ptr != NULL)
			    ptr[0] = '\0';
			/* XXX : add a # before chan_name if not specified ! */
			chan_name = pmalloc(s->p, strlen(address));
			strcpy(chan_name,address);
		    }

		    /* temp store channel name in aux1 */
		    jp->aux1 = (void*) s;
		    s->aux1 = (void *) chan_name;
		    /* check that room does not already exist
		     * before going into next phase */
		    mio_reset(s->m, irc_parse_phase_zero, arg);
		    mio_write(s->m, NULL, spools(s->p, "MODE ", chan_name, "\n", s->p), -1);
		}
                return;
	    }
	}

	/* if there is no session for this user, send an error */
	/* query requests on conference ns need to be registered */
	if(s == NULL) {
	    jutil_error(jp->x, TERROR_REGISTER);
	    xmlnode_insert_cdata(xmlnode_get_tag(jp->x, "error"), "You must be registred to a server", -1);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	if ((strncmp(xmlnode_get_name(jp->iq),"query",5) == 0) 
	    && (jpacket_subtype(jp) == JPACKET__SET)) {

	    /* 
	     * for clients that translate commands like /nick, /topic
	     *     to iq query requests
	     * by BO, 2001/10/16
	     */

	    /* XXX only nickname change is currently implemented 
	     *     add what is possible (passwd, name, privacy)
	     * by BO, 2001/09/05
	     */
	    if (xmlnode_get_tag_data(jp->iq, "nickname") == NULL) {
		jutil_error(jp->x, TERROR_NOTIMPL);
		deliver(dpacket_new(jp->x), NULL);
		return;
	    }

	    s->aux1 = (void*)jp;
	    mio_write(s->m, NULL, spools(jp->p, "NICK :", xmlnode_get_tag_data(jp->iq, "nickname"), "\n", jp->p), -1);
	    xmlnode_hide(xmlnode_get_tag(jp->iq, "nickname"));
	    return;
	} 

	/* XXX : add handling of destroy requests */

        /* we don't accept this */
        jutil_error(jp->x, TERROR_BAD);
        deliver(dpacket_new(jp->x), NULL);
        return;
    }

    /* from now, user must have a session */
    if (s == NULL) {
	jutil_error(jp->x, TERROR_REGISTER);
	deliver(dpacket_new(jp->x),NULL);
	return;
    }

    if(jp->type == JPACKET_PRESENCE && jpacket_subtype(jp) == JPACKET__UNAVAILABLE)
    {
	char *chan_name = irc_get_channel(jp->p, jid_full(jp->to));

        /* end session */
        log_debug(ZONE, "[irc_parser] Initial session shut down for user %s for channel %s on server %s", jid_full(jp->from), chan_name, s->server);
	
	/* if a channel name is provided, just leave this channel
	 * either, leave the IRC server */
	if (chan_name != NULL) {
	    chans channel = xhash_get(s->channels, make_lowercase(chan_name));
	    if (channel == NULL) {
		log_error(ZONE,"[irc_parser] channel not found in channels list !");
	    } else {
		/* remove channel from the session channels list */
		xhash_zap(s->channels, channel->id);
		xhash_zap(s->channels, channel->name);
		xhash_free(channel->users);
		pool_free(channel->p);
	    }
	    mio_write(s->m, NULL, spools(jp->p, "PART ", chan_name, "\n", jp->p), -1);
	} else {
	    mio_write(s->m, NULL, "QUIT", -1);
	    /* session is removed when receiving the MIO_CLOSED */
	    mio_close(s->m);
	}
        return;
    }
    else if(jp->type == JPACKET_PRESENCE)
    {
        xmlnode_free(jp->x);
        return;
    }

    /* all that's left are messages to the group */
    if(jp->type == JPACKET_MESSAGE)
    {
        char *msg;
        chans channel = xhash_get(s->channels, spools(jp->p, jp->to->user, "@", jp->to->server, jp->p));

        log_debug(ZONE, "[irc_parser] Got Message to %s, channel found %X", 
		  jid_full(jp->to), channel);

        if(channel == NULL)
        {
            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        log_debug(ZONE, "[irc_parser] Channel object says: %s : %s", channel->name, channel->id);

        if(jp->to->resource != NULL) /* private message */
        {
            jutil_error(jp->x, TERROR_NOTIMPL);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

	/* topic change request */
	if (xmlnode_get_tag(jp->x,"subject") != NULL) {
	    char *topic = UTF8_to_str(jp->p, xmlnode_get_data(xmlnode_get_tag(jp->x, "subject")));
	    mio_write(s->m, NULL, spools(jp->p, "TOPIC ", irc_get_channel(jp->p, jid_full(jp->to)), " :", topic, "\n", jp->p), -1);
	    xmlnode_free(jp->x);
	    return;
	}

        msg = UTF8_to_str(jp->p, xmlnode_get_data(xmlnode_get_tag(jp->x, "body")));

        /* XXX break up messages bigger than 450 chars.. don't chop them */
	if (msg == NULL)
	    return;
        if(strlen(msg) > 450)
            msg[450] = '\0';

        
        /*
        * this if block handles the packet.. 
        * each section is responsible for freeing the packet in it's own way
        */
        if(strncasecmp(msg, "/me ", 4) == 0)
        {
            /* this should be an action */
            jutil_tofrom(jp->x);
            msg += 4;
            xmlnode_put_attrib(jp->x, "from", channel->nick_id);
            jpacket_reset(jp);

            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :\001ACTION ", msg, "\001\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
	else if(strncasecmp(msg, "/nick ", 6) == 0)
	{
	    msg += 6;
	    mio_write(s->m, NULL, spools(jp->p, "NICK :", msg, "\n", jp->p), -1);
	    xmlnode_free(jp->x);
	}
        else if(strncasecmp(msg, "/say ", 5) == 0)
        {
            jutil_tofrom(jp->x);
            msg += 5;
            xmlnode_put_attrib(jp->x, "from", channel->nick_id);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), msg, -1);

            jpacket_reset(jp);
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/raw ", 5) == 0)
        {
            msg += 5;
            mio_write(s->m, NULL, spools(jp->p, msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg,"/whowas ",7)==0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOWAS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/whois ", 7) == 0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOIS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/msg ", 5) == 0 || strncasecmp(msg, "/query ", 7) == 0)
        {
            char *nick;
            if(strncasecmp(msg, "/msg ", 5) == 0)
                msg += 5;
            else
                msg += 7;

            if(msg != NULL && msg[0] != '\0')
            {
                nick = msg;
                msg = strchr(msg,' ');
                if(msg != NULL)
                {
                    msg[0] = '\0';
                    msg++;
                    /* strip the @ and + from the user */
                    if(nick[0] == '@' || nick[0] == '+') nick++;
                    mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", nick, " :", msg, "\n", jp->p), -1);
                }
            }
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), ">>> PRIVMSG SENT", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/mode ", 6) == 0)
        {
            msg += 6;
            mio_write(s->m, NULL, spools(jp->p, "MODE ", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
	else if (strncasecmp(msg, "/topic ", 8) == 0)
	{
	    msg += 8;
	    mio_write(s->m, NULL, spools(jp->p, "TOPIC ", irc_get_channel(jp->p, jid_full(jp->to)), " :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
	}
        else if (strncasecmp(msg, "/leave ", 7) == 0)
	{
	    msg += 7;
	    mio_write(s->m, NULL, spools(jp->p, "PART ", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
	}
	else if(msg[0] == '/')
        {
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), "/ COMMAND NOT SUPPORTED", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else /* standard message */
        {
            jutil_tofrom(jp->x);
            xmlnode_put_attrib(jp->x, "from", channel->nick_id);
            jpacket_reset(jp);
            log_debug(ZONE, "[irc_parser] sending %s", spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p));
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        return;
    }

    /* dunno what this is */
    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x), NULL);
}
